
#pragma semicolon 1
#pragma tabsize 0
#pragma newdecls required
#include <sourcemod>
#include <pan0s>

#define CVAR_FLAGS			FCVAR_NOTIFY
#define VERSION             "v1.0"
#define VOTE_NO             "###no###"
#define VOTE_YES            "###yes###"
#define CMD                 "sm_ivote_type"

enum struct Player
{
    int id;
    char name[64];
    char steamId[32];
}

enum struct VoteInfo
{
    int type;
    char value[64];
    Player target;
    Player creator;
}

enum
{
    TYPE_MAP    =0,
    TYPE_KICK   =1,
    TYPE_BAN    =2,
    TYPE_SLAY   =3,
}

// ConVar
ConVar cvar_passPercentage;
ConVar cvar_bot_included;
ConVar cvar_self_included;
ConVar cvar_team_filter;
ConVar cvar_bot_filter;
ConVar cvar_vote_cd;
ConVar cvar_ivotes_on[4];

ArrayList g_listOn;
ArrayList g_listClients[MAXPLAYERS+1];
ArrayList g_listSteamId[MAXPLAYERS+1];

// Menu g_mMenu[MAXPLAYERS+1];
int g_iType[MAXPLAYERS+1];
float g_fCD[MAXPLAYERS+1];

VoteInfo g_vInfo;

Handle g_tResetTimer;

public Plugin myinfo =
{
	name = "L4D2 IVote",
	description = "A simple plugin to allow players imitate sourcemod vote.",
	author = "pan0s",
	url = ""
};

public void OnPluginStart()
{
    LoadTranslations("l4d2_tags.phrases");
    LoadTranslations("l4d2_ivote.phrases");
    LoadTranslations("common.phrases");
    
    RegConsoleCmd("sm_ivote", HandleCmdVoteMenu);
    RegConsoleCmd("sm_tiv", GeHandleCmdTiv);
    RegConsoleCmd("sm_ivoteblock", HandleCmdVoteBlockMenu);
    RegConsoleCmd(CMD, HandleCmdIVote);

    CreateConVar("l4d2_ivote_version", VERSION, "L4D2 ivote version", CVAR_FLAGS);

    cvar_ivotes_on[0] = CreateConVar("l4d2_ivoteMapOn", "0", "(NOT FINISHED) Is vote map on?\n0=Off, 1=On", CVAR_FLAGS, true, 0.0, true, 0.0);
    cvar_ivotes_on[1] = CreateConVar("l4d2_ivoteKickOn", "1", "Is vote kick on?\n0=Off, 1=On", CVAR_FLAGS, true, 0.0, true, 1.0);
    cvar_ivotes_on[2] = CreateConVar("l4d2_ivoteBanOn", "1", "Is vote ban (15min.) on?\n0=Off, 1=On", CVAR_FLAGS, true, 0.0, true, 1.0);
    cvar_ivotes_on[3] = CreateConVar("l4d2_ivoteSlayOn", "1", "Is vote slay on?\n0=Off, 1=On", CVAR_FLAGS, true, 0.0, true, 1.0);
    cvar_passPercentage = CreateConVar("l4d2_passPercentage", "0.60", "percent required for successful kick vote", CVAR_FLAGS, true, 0.00, true, 1.00);
    cvar_bot_included = CreateConVar("l4d2_botIncluded", "1", "Is vote slay on?\n0=Off, 1=On but not included in kick and ban vote list. 2=On all.", CVAR_FLAGS, true, 0.0, true, 2.0);
    cvar_team_filter = CreateConVar("l4d2_teamFilter", "1", "Will different team players not be included in the players list?\n0=Off, 1=On", CVAR_FLAGS, true, 0.0, true, 1.0);
    cvar_bot_filter = CreateConVar("l4d2_botFilter", "2", "Filter BOT in the players list (bot_included > 0 only)\n0=Off, 1=Survivor, 2=Infected", CVAR_FLAGS, true, 0.0, true, 2.0);
    cvar_self_included = CreateConVar("l4d2_selfIncluded", "1", "Is vote slay on?\n0=Off, 1=On", CVAR_FLAGS, true, 0.0, true, 1.0);
    cvar_vote_cd = CreateConVar("l4d2_vote_cd", "30.0", "What is the cool down time of each vote for every player?", CVAR_FLAGS, true, 0.0);

    AutoExecConfig(true, "l4d2_ivote");

    g_listOn = new ArrayList();
    for(int i=i; i<=MaxClients; i++)
    {
        g_listClients[i] = new ArrayList();
        g_listSteamId[i] = new ArrayList(ByteCountToCells(32));
    }

    UpdateOnList();
    g_vInfo.type = -1;
}

public void UpdateOnList()
{
    g_listOn.Clear();
    for(int i=0; i<sizeof(cvar_ivotes_on); i++)
        if(cvar_ivotes_on[i].BoolValue) 
            g_listOn.Push(i);
}

public void GetIVoteCmd(char[] buffer, const int size, const int type)
{
    Format(buffer, size, "%s %d", CMD, type);
}

public Action HandleCmdVoteMenu(int client, int args)
{
    Menu menu = new Menu(HandleVoteMenuOpen);
    menu.SetTitle("%T", "MENU", client);
    
    for(int i=0; i<g_listOn.Length; i++)
    {
        char cmd [24];
        char option [64];
        char optionName [10];
        GetIVoteCmd(cmd, sizeof(cmd), g_listOn.Get(i));

        if(TranslationPhraseExists(cmd)) Format(option, sizeof(option), "%T",cmd, client);
        else Format(option, sizeof(option), "%s", cmd);
        Format(optionName, sizeof(optionName),"option%d", i);
        menu.AddItem(optionName, option);
    }
    menu.ExitButton = true;
    menu.Display(client, MENU_TIME_FOREVER);
}

public Action GeHandleCmdTiv(int client, int args)
{
    Menu mVote  = new Menu(Handler_VoteCallback, MENU_ACTIONS_ALL);
    g_iType[client] = TYPE_SLAY;
    g_vInfo.type = TYPE_SLAY;
    UpdateVoteInfo(client, client);

    for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
    {
        char translatedCmd[64];
        Format(translatedCmd, sizeof(translatedCmd), "%T", "sm_ivote_type 3", i);
        CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "START_VOTE", i, client, translatedCmd, g_vInfo.target.name);
    }

    StartVote(client, mVote);
}

public Action HandleCmdVoteBlockMenu(int client, int args)
{
	Panel panel = new Panel();
	panel.DrawText("Init vote...");

	char line[128];
	Format(line, sizeof(line), "Close");
	panel.DrawItem(line, ITEMDRAW_DEFAULT);

	panel.Send(client, HandlePanelSelected, MENU_TIME_FOREVER);
	delete panel;
}

public int HandlePanelSelected(Menu menu, MenuAction action, int client, int selectedIndex)
{
	switch(action) 
	{
		case MenuAction_End: delete menu;
	}
	return 0;
}

public int HandleVoteMenuOpen(Handle menu, MenuAction action, int client, int option) 
{
    switch(action)
    {
        case MenuAction_End: delete menu;
        case MenuAction_Select:
        {
            if(IsValidClient(client))
            {
                char cmd[24];
                GetIVoteCmd(cmd, sizeof(cmd), g_listOn.Get(option));
                FakeClientCommand(client, cmd);
            }
        }
    }
}

public void UpdatePlayerList(int client)
{
    g_listClients[client].Clear();
    g_listSteamId[client].Clear();
    int team = GetClientTeam(client);
    for(int i=1; i<=MaxClients; i++)
    {
        // Checking--------------------->
        // Common
        if(!IsValidClient(i) || 
            (i == client && !cvar_self_included.BoolValue) || 
            (g_iType[client] == TYPE_SLAY && !IsPlayerAlive(i))
        ) continue;

        int targetTeam = GetClientTeam(i);

        // BOT & Team check
        if(IsFakeClient(i))
        {
            if(cvar_team_filter.BoolValue && team != targetTeam) continue;
            if(cvar_bot_included.IntValue == 0) continue;
            if(cvar_bot_included.IntValue == 1 && IsFakeClient(i) && (g_iType[client] == TYPE_KICK || g_iType[client] == TYPE_BAN)) continue;
            if(cvar_bot_filter.IntValue >0)
            {
                if(cvar_bot_filter.IntValue == 1 && targetTeam == TEAM_SURVIVOR) continue;
                if(cvar_bot_filter.IntValue == 2 && targetTeam == TEAM_INFECTED) continue;
            }
        }
        // <--------------------Checking

        char steamId[32];
        GetClientAuthId(i, AuthId_Steam2, steamId, 32);
        g_listClients[client].Push(i);
        g_listSteamId[client].PushString(steamId);
    }
}

public Action HandleCmdIVote(int client, int args)
{
    if(args != 1) return Plugin_Handled;

    if(g_vInfo.type != -1)
    {
        CPrintToChat(client, "%T%T", "TAG_NOTICE", client, "IS_VOTING", client);
        return Plugin_Handled;
    }

    if(g_fCD[client] > 0.0)
    {
        CPrintToChat(client, "%T%T", "TAG_NOTICE", client, "PLEASE_WAIT", client, g_fCD[client]);
        return Plugin_Handled;
    }

    char sIndex[2];
    GetCmdArgString(sIndex, sizeof(sIndex));
    g_iType[client] = StringToInt(sIndex);

    UpdatePlayerList(client);
    MenuChoosePlayer(client);
    
    return Plugin_Handled;
}

public void MenuChoosePlayer(int client)
{
    if(g_listClients[client].Length == 0)
    {
        CPrintToChat(client, "%T%T", "TAG_NOTICE", client, "NO_PLAYER", client);
        return;
    }

    char cmd[24];
    GetIVoteCmd(cmd, sizeof(cmd), g_iType[client]);

    Menu menu;

    switch(g_iType[client])
    {
        case TYPE_MAP: menu = new Menu(HandleMenuVoteMap);
        case TYPE_KICK, TYPE_BAN, TYPE_SLAY: menu = new Menu(HandleMenuVoteCommon);
    }

    menu.SetTitle("%T: ", cmd, client);

    if(g_iType[client] != TYPE_MAP)
    {
        for(int i=0; i<g_listClients[client].Length; i++)
        {
            int target = g_listClients[client].Get(i);
            bool isBot = IsFakeClient(target);
            char option[64];
            Format(option, sizeof(option), "%N% %s", target, isBot? "(BOT)":"");
            menu.AddItem("", option);
        }
    }


    menu.ExitBackButton = true;
    menu.ExitButton = true;
    menu.Display(client, MENU_TIME_FOREVER);
}

public bool AuthResuilt()
{
    char steamId[32];
    GetClientAuthId(g_vInfo.target.id, AuthId_Steam2, steamId, sizeof(steamId));
 
    if(StrEqual(steamId, g_vInfo.target.steamId, true)) return true;

    return false;
}

public bool AuthTarget(int client, int selected, int target)
{
    char steamId[32];
    GetClientAuthId(target, AuthId_Steam2, steamId, sizeof(steamId));

    char selectedSteamId[32];
    g_listSteamId[client].GetString(selected, selectedSteamId, sizeof(selectedSteamId));

    if(StrEqual(steamId, selectedSteamId, true)) return true;

    CPrintToChat(client, "%T%T", "TAG_NOTICE", client, "AUTH_FAIL", client, target);
    return false;
}

public int Handler_VoteCallback(Menu menu, MenuAction action, int client, int param2)
{
    switch(action)
    {
        case MenuAction_End: delete menu;
        case MenuAction_Display: 
        {
            if(g_vInfo.type != TYPE_MAP)
            {
                char title[64];
                menu.GetTitle(title, sizeof(title));

                Format(title, sizeof(title), "QUESTION_%s", title);

                char buffer[255];
                Format(buffer, sizeof(buffer), "%T", title, client, g_vInfo.target.name);

                Panel panel = view_as<Panel>(param2);
                panel.SetTitle(buffer);
            }

        }
        case MenuAction_DisplayItem:
        {
            char display[64];
            menu.GetItem(param2, "", 0, _, display, sizeof(display));
        
            if (strcmp(display, "VOTE_NO") == 0 || strcmp(display, "VOTE_YES") == 0)
            {
                char buffer[255];
                Format(buffer, sizeof(buffer), "%T", display, client);

                return RedrawMenuItem(buffer);
            }
        }
        case MenuAction_Select:
        {
            if(g_vInfo.type != TYPE_MAP)
            {
                for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
                {
                    char vote[32];
                    if(param2 == 0) Format(vote, sizeof(vote), "{GREEN}%T{DEFAULT}", "VOTE_YES", i);
                    else Format(vote, sizeof(vote), "{ORANGE}%T{DEFAULT}", "VOTE_NO", i);
                    CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "VOTED", i, client, vote);
                }
            }
        }
        case MenuAction_VoteCancel, VoteCancel_NoVotes:
        {
            if(IsValidClient(client)) CPrintToChat(client, "%T%T", "TAG_NOTICE", client, "No Votes Cast", client);
            g_vInfo.type = -1;
        }
        case MenuAction_VoteEnd:
        {
            char item[PLATFORM_MAX_PATH], display[64];
            float percent, limit;
            int votes, totalVotes;

            GetMenuVoteInfo(param2, votes, totalVotes);
            menu.GetItem(client, item, sizeof(item), _, display, sizeof(display));
            
            if (strcmp(item, VOTE_NO) == 0 && client == 1)
            {
                votes = totalVotes - votes; // Reverse the votes to be in relation to the Yes option.
            }
            
            percent = GetVotePercent(votes, totalVotes);
            
            limit = cvar_passPercentage.FloatValue;
            
            // A multi-argument vote is "always successful", but have to check if its a Yes/No vote.
            if ((strcmp(item, VOTE_YES) == 0 && FloatCompare(percent,limit) < 0 && client == 0) || (strcmp(item, VOTE_NO) == 0 && client == 1))
            {
                /* :TODO: g_voteTarget should be used here and set to -1 if not applicable.
                */
                LogAction(-1, -1, "Vote failed.");
                for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
                    CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "VOTE_FAILED", i,RoundToNearest(100.0*limit), RoundToNearest(100.0*percent), totalVotes);
            }
            else
            {
                for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
                    CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "VOTE_SUCCESSFUL", i, RoundToNearest(100.0*percent), totalVotes);
                
                if(g_vInfo.type != TYPE_MAP)
                {
                    int target = g_vInfo.target.id;
                    switch (g_vInfo.type)
                    {
                        case TYPE_KICK:
                        {
                            if(IsValidClient(g_vInfo.target.id))
                            {
                                if(AuthResuilt())
                                {
                                    LogAction(-1, -1, "[IVOTE]: The kick vote creadted by \"%s<%s><>\" passed (%.0f%% of %d voted). Kicked \"%s<%s><>\"", g_vInfo.creator.name, g_vInfo.creator.steamId, percent*100, totalVotes, g_vInfo.target.name, g_vInfo.target.steamId);
                                    KickClient(g_vInfo.target.id, "%T", "KICKED_MSG", g_vInfo.target.id);
                                    for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
                                        CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "KICKED", i, g_vInfo.target.name);
                                }
                                else
                                {
                                    LogAction(-1, -1, "[IVOTE]: The kick vote creadted by \"%s<%s><>\" passed (%.0f%% of %d voted). Failed to kick \"%s<%s><>\", client already disconnected.", g_vInfo.creator.name, g_vInfo.creator.steamId, percent*100, totalVotes, g_vInfo.target.name, g_vInfo.target.steamId);
                                    for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
                                        CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "FAILED_TO_KICK", i, g_vInfo.target.name);
                                }
                            }
                            else
                            {
                                LogAction(-1, -1, "[IVOTE]: The kick vote creadted by \"%s<%s><>\" passed (%.0f%% of %d voted). Failed to kick \"%s<%s><>\", client is invaild", g_vInfo.creator.name, g_vInfo.creator.steamId, percent*100, totalVotes, g_vInfo.target.name, g_vInfo.target.steamId);
                            }
                        }
                            
                        case TYPE_BAN:
                        {
                            char banCmd[512];
                            Format(banCmd, sizeof(banCmd), "sm_addban 15 %s \"%T\"", g_vInfo.target.steamId, "BANNED_MSG", g_vInfo.target.id);
                            PrintToServer(banCmd);
                            ServerCommand(banCmd);
                            LogAction(-1, -1, "[IVOTE]: The ban vote creadted by \"%s<%s><>\" passed (%.0f%% of %d voted). Banned \"%s<%s><>\" for 15 minutes.", g_vInfo.creator.name, g_vInfo.creator.steamId, percent*100, totalVotes, g_vInfo.target.name, g_vInfo.target.steamId);
                            for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
                                CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "BANNED", i, g_vInfo.target.name);
                            if(AuthResuilt())
                                    KickClient(g_vInfo.target.id, "%T", "BANNED_MSG", g_vInfo.target.id);
                        }

                        case TYPE_SLAY:
                        {
                            if(IsValidClient(target))
                            {
                                if(IsPlayerAlive(target))
                                {
                                    ForcePlayerSuicide(target);
                                    LogAction(-1, -1, "[IVOTE]: The slay vote of created by \"%s<%s><>\" passed (%.0f%% of %d voted). Slayed \"%s<%s><>\"", g_vInfo.creator.name, g_vInfo.creator.steamId, percent*100, totalVotes, g_vInfo.target.name, g_vInfo.target.steamId);
                                    for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
                                        CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "SLAYED", i, g_vInfo.target.name);
                                }
                                else
                                {
                                    LogAction(-1, -1, "[IVOTE]: The slay vote of \"%s<%s><>\" passed (%.0f%% of %d voted). Failed to slay \"%s<%s><>\", target is dead.", g_vInfo.creator.name, g_vInfo.creator.steamId, g_vInfo.target.name, g_vInfo.target.steamId);
                                    for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
                                        CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "FAILED_TO_SLAY", i, g_vInfo.target.name);
                                }
                            }
                            else
                            {
                                LogAction(-1, -1, "[IVOTE]: The slay vote of \"%s<%s><>\" passed (%.0f%% of %d voted). Failed to slay \"%s<%s><>\", target is invaild.", g_vInfo.creator.name, g_vInfo.creator.steamId, g_vInfo.target.name, g_vInfo.target.steamId);
                            }
                        }
                    }
                }
                else
                {
                    //TODO: change map
                }
            }
            ResetVoteInfo();
        }
    }

	return 0;
}

public void ResetVoteInfo()
{
    g_vInfo.type = -1;
    g_vInfo.creator.id = -1;
    g_vInfo.target.id = -1;

    for(int i=0; i<64; i++)
    {
        g_vInfo.creator.name[i] = '\0';
        g_vInfo.target.name[i] = '\0';
    }

    for(int i=0; i<32; i++)
    {
        g_vInfo.creator.steamId[i] = '\0';
        g_vInfo.target.steamId[i] = '\0';
    }

}

public Action HandleTimerCD(Handle timer, int client)
{
    g_fCD[client] -= 1.0;
    if(g_fCD[client] > 0.0)
        CreateTimer(1.0, HandleTimerCD, client);
}

public void StartVote(int client, Menu menu)
{
    // Make a panel to announc
    for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
        FakeClientCommand(i, "sm_ivoteblock");

    char cmd[24];
    GetIVoteCmd(cmd, sizeof(cmd), g_iType[client]);
    menu.SetTitle(cmd);
    menu.AddItem(VOTE_YES, "VOTE_YES");
    menu.AddItem(VOTE_NO, "VOTE_NO");
    menu.ExitButton = false;
    menu.DisplayVoteToAll(20);

    g_fCD[client] = cvar_vote_cd.FloatValue;
    CreateTimer(1.0, HandleTimerCD, client);
}

public void UpdateVoteInfo(int client, int target)
{
    g_vInfo.creator.id = client;
    GetClientName(client, g_vInfo.creator.name, 64);
    GetClientAuthId(client, AuthId_Steam2, g_vInfo.creator.steamId, 32);

    g_vInfo.target.id = target;
    GetClientName(target, g_vInfo.target.name, 64);
    GetClientAuthId(target, AuthId_Steam2, g_vInfo.target.steamId, 32);
}

public int HandleMenuVoteMap(Handle menu, MenuAction action, int client, int option) 
{
    switch(action)
    {
        case MenuAction_End: delete menu;
        case MenuAction_Select:
        {

        }
        case MenuAction_Cancel: 
		{
			if(option == MenuCancel_ExitBack) 
			{
				FakeClientCommand(client, "sm_ivote");
			}
		}
    }
}

public int HandleMenuVoteCommon(Handle menu, MenuAction action, int client, int option) 
{
    switch(action)
    {
        case MenuAction_End: delete menu;
        case MenuAction_Select:
        {
            UpdateVoteInfo(client, g_listClients[client].Get(option));

            if(!AuthTarget(client, option, g_vInfo.target.id)) return;

            char cmd[24];
            GetIVoteCmd(cmd, sizeof(cmd), g_iType[client]);

            for(int i=1; i<=MaxClients; i++) if(IsValidClient(i))
            {
                char translatedCmd[64];
                Format(translatedCmd, sizeof(translatedCmd), "%T", cmd, i);
                CPrintToChat(i, "%T%T", "TAG_ANNOUNCE", i, "START_VOTE", i, client, translatedCmd, g_vInfo.target.name);
            }
            Menu mVote  = new Menu(Handler_VoteCallback, MENU_ACTIONS_ALL);
            g_vInfo.type = g_iType[client];
            
            StartVote(client, mVote);

            //Create timer to reset vote info. just in case
            delete g_tResetTimer;
            g_tResetTimer = CreateTimer(180.0, HandleTimerResetVoteInfo);
        }
        case MenuAction_Cancel: 
		{
			if(option == MenuCancel_ExitBack) 
			{
				FakeClientCommand(client, "sm_ivote");
			}
		}
    }
}

public Action HandleTimerResetVoteInfo(Handle timer)
{
    ResetVoteInfo();
    g_tResetTimer = null;
}


float GetVotePercent(int votes, int totalVotes)
{
	return float(votes) / float(totalVotes);
}